/*----------------------------------------------------------------------------
--
--  checker-err.c - This is the main program for the Checker-ERR server.
--  It receives  DNS packets from a tcp port and pile the data to Steve's 
--  code for analyzing.
--
--  Copyright (c) 1994, 1995
--                University of Southern California, Los Angeles, CA
--                All rights reserved.
--
--  Redistribution and use of this software in source or binary form is
--  permitted given that the above copyright notice and this paragraph are
--  duplicated in all such forms and that any documentation and material
--  related to such distribution and use acknowledge that the software
--  was developed by University of Southern California.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMIT TO, THE IMPLIED
--  WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--  THE UNIVERSITY AND THE AUTHORS ASSUME NO RESPONSIBILITY FOR DAMAGES
--  OF ANY KIND RESULTING FROM THE USE OF THIS SOFTWARE.
--
--  Please contact Dr. Peter Danzig of Computer Science Department at
--        danzig@usc.edu for further information.
--
--
===========================================================================
--                           UPDATE HISTORY                              --
--                                                                       --
-- VER     Date         Explanation of Changes              Author       --
---------------------------------------------------------------------------
--  0   |  5/94    | initial version                | Charles Chan - USC --
--  1   |  6/95    | fixed code formatting          | Steve Miller - USC --
-------------------------------------------------------------------------*/

#include  <stdio.h>
#include  <time.h>
#include  <sys/types.h>
#include  <sys/socket.h>
#include  <netinet/in.h>
#include  <arpa/inet.h>
#include  <fcntl.h>
#include  <signal.h>
#include  <sys/ioctl.h>
#include  <sys/wait.h>
#include  "checker.h"

SOURCE_INFO   src;
QUERY_PACKET  qp;
int           runDaemon = 0;
char         *pname;

u_long        totalQueryCount = 0;
u_long        totalReponseCount = 0;
u_long        totalErrorQueries = 0;
u_long        lastTotalQueryCount = 0;
u_long        lastTotalReponseCount = 0;
u_long        lastTotalErrorQueries = 0;
time_t        programStartTime = 0;

extern FILE *tfp;
extern char base_password[50];
extern u_long password_check;
extern u_long packet_tracing;


/*-----------------------------------------------------------------------------
--
--  main() - Entry point for the checker database processor
--
------------------------------------------------------------------------------*/

main(argc, argv)
int  argc;
char *argv[];
    {
    u_short  port;
    int sockfd, newsockfd, clilen;
    struct sockaddr_in cli_addr, serv_addr;

    /* Check our input argument count. */

    pname = argv[0];
    if (argc < 2) {
        printf("usage: %s <port number> [-daemon]\n", pname);
        exit(0); }

    /*
     * Read the incoming connection accept port number, and check that it
     * is in a valid range.
     */

    sscanf(argv[1], "%hd", &port);
    if (port < 5000) {
        printf("usage: %s <port number> [-daemon]\n", pname);
        printf("%s: please take a port number larger than 5000.\n", pname);
        exit(0);
        }

    /*
     * Check to see if the user added another parameter requesting
     * that this program run in the background as a daemon process.
     */

    if (argc == 3) {
        if (!strcmp(argv[2], "-daemon")) {
            printf("%s: runs as a daemon process\n", argv[0]);
            runDaemon = 1; }
        else {
            printf("usage: %s <port number> [-daemon]\n", pname);
            printf("%s: unknown option %s.\n", pname, argv[2]);
            exit(0); }
        }

    /*
     * Read the control configuration that decides many of the
     * reporting and error checking parameters (as well as some of the
     * path names and other maintenance functions).
     */

    read_configuration();

    /*
     * If requested then restart ourselves as a daemon process and 
     * relinquish terminal control back to the user. 
     */

    if (runDaemon) StartDaemon();

    /*
     * Open a TCP socket (an Internet stream socket).
     */

    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        fprintf(stderr, "server: can't open stream socket");

    /*
     * Bind our local address so that the client can send to us.
     */

    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_addr.sin_port = htons(port);

    if (bind(sockfd, (struct sockaddr *) & serv_addr, sizeof(serv_addr)) < 0)
        fprintf(stderr, "server: can't bind local address");

    /*
     * Wait for a connection from a client process
     */

    if (!runDaemon) printf("waiting for connection....\n");
    listen(sockfd, 1);
    clilen = sizeof(cli_addr);
    newsockfd = accept(sockfd, (struct sockaddr *) & cli_addr, &clilen);
    if (newsockfd < 0) fprintf(stderr, "server: accept error");
    if (!runDaemon) printf("connection established....\n");

    /*
     * Loop forever reading data from the connection.  When the socket
     * breaks return here, clean up this end, and exit the program.
     */

    ReadSocket(newsockfd);
    close(newsockfd);
    }


/*-----------------------------------------------------------------------
--
--  itoc - returns inet address in dot format given HEX address
--
-----------------------------------------------------------------------*/
char *
itoc(str, len, rlt)
char *str;
int   len;
char *rlt;
    {
    int             i;
    char            asc[8];

    rlt[0] = '\0';

    for (i = 0; i < len; i++) {
        sprintf(asc, "%d.", (u_char) str[i]);
        strcat(rlt, asc);
        }

    rlt[strlen(rlt) - 1] = '\0';
    return (rlt);
    }


/*-----------------------------------------------------------------------
--
--  ChildExitHandle()
--
-----------------------------------------------------------------------*/
void 
ChildExitHandle()
    {
    union wait      status;
    while (wait3(&status, WNOHANG, 0) > 0);
    }


/*-----------------------------------------------------------------------
--
--  ReadNBytes - reads the given n bytes of data from fd and puts into
--  the given buffer *ptr.
--
-----------------------------------------------------------------------*/
int
ReadNBytes(fd, ptr, nbytes)
register int    fd;
register char  *ptr;
register int    nbytes;
    {
    int nleft, nread;

    nleft = nbytes;

    while (nleft > 0) {

        nread = read(fd, ptr, nleft);

        if (nread < 0) {
            if( nleft == nbytes ) return( nread );
            else return( nbytes - nleft );
            }

        else if (nread == 0) break;

        nleft -= nread;
        ptr += nread;
        }

    return (nbytes - nleft);    /* return >= 0 */
    }


/*-----------------------------------------------------------------------
--
--  ReadSocket - reads DNS packets from sockfd and prepares for passing to 
--  analyze. 
--
-----------------------------------------------------------------------*/
ReadSocket(sockfd)
int sockfd;
    {
    int  bytes_read = 1;
    int  nsrc = sizeof(SOURCE_INFO);
    int  nqp = sizeof(QUERY_PACKET);
    int  command_number;
    char tnam[30];
    char noError = 1;

    static char    *rcode_string[] = {
        "No error",
        "Query format error",
        "Server failure",
        "No such name is found",
        "Query not supported",
        "Request Refused By Server" };

    /* Remember starting time of the program */
    
    programStartTime = time ( (time_t *)0 );

    while (bytes_read > 0) {

        bytes_read = ReadNBytes(sockfd, (char *) &src, nsrc);

        if (bytes_read != nsrc && bytes_read > 0) {
            fprintf(stderr, "ReadSocket: ReadNBytes() error\n");
            noError = 0; }

        bytes_read = ReadNBytes(sockfd, (char *) &qp, nqp);

        if (bytes_read != nqp && bytes_read > 0) {
            fprintf(stderr, "ReadSocket: ReadNBytes() error\n");
            noError = 0; }

        /*
         * Look for control commands.  The primary ones are 1, which
         * causes us to reread our configuration file; and 5 which causes
         * the report processor to run.
         */

        if( strncmp( qp.qname, "checker.control.command.", 24 ) == 0 ) {

            if( qp.type == 0 ) {

                if( password_check && validate_password(qp.qname) );
                else {

                    sscanf( &qp.qname[24], "%d", &command_number );

                    switch( command_number ) {
                        case  1: read_configuration(); break;
                        case  5: dump_tables(); break;
                        default: break;
                        }
                    }
                }
            }

        /*
         * If we are not dealing with a control command then record the
         * packet and pass it on to the actual statistical checking
         * code.  We currently do not put queries captured by the PEC
         * system into the database, but we do count them.
         */

        else {

            if( packet_tracing && tfp != NULL ) {
                fwrite( &src, 1, sizeof(src), tfp );
                fwrite( &qp,  1, sizeof(qp),  tfp );
                fflush( tfp ); }

            if( strncmp(qp.qname, "** PEC captured", 14) != 0 )
                handle_packet(&qp, &src);

            if (qp.type == 0)      totalQueryCount++;
            else {                 totalReponseCount++;
                if (qp.rcode != 0) totalErrorQueries++; }
            }
        }
    }


/*-----------------------------------------------------------------------------
--
--  StartDaemon() - to start this program as a daemon process.
--
------------------------------------------------------------------------------*/
StartDaemon()
    {
    int             fd;
    int             childPid;

    if (getppid() == 1) return;

    /*
     * Ignore the terminal stop signals (BSD).
     */

#ifdef SIGTTOU
    signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
    signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
    signal(SIGTSTP, SIG_IGN);
#endif

    /* let exit the parent */

    if ((childPid = fork()) < 0)
        printf("can't fork first child in starting daemon\n");
    else if (childPid > 0)
        exit(0); /* parent now is gone */

    /* change process group */

#ifdef  SIGTSTP            /* BSD */

    if (setpgrp(0, getpid()) == -1)
        printf("can't change process group in starting daemon\n");

    /* we don't want the terminal association */

    if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
        ioctl(fd, TIOCNOTTY, (char *) NULL);    /* lose controlling tty */
        close(fd); }
    signal(SIGCHLD, ChildExitHandle);

#else                /* System V */

    if (setpgrp() == -1)
        err_sys("can't change process group");

    signal(SIGHUP, SIG_IGN);    /* immune from pgrp leader death */

    if ((childpid = fork()) < 0)
        printf("can't fork second child in starting daemon\n");
    else if (childpid > 0)
        exit(0);        /* first child */

#endif

    }


/*-----------------------------------------------------------------------
--
-- int validate_password()
--     A little routine to check that the password on a control command
--     is correct.
--
--  GIVEN          : 
--  GLOBAL USED    : 
--  RETURNS        : 
--
--  POST CONDITIONS: 
--
--  COMMENTS:
--      none
--
-----------------------------------------------------------------------*/
int
validate_password(str)
char   *str;
    {
    int   i, j;
    char  tstr[50];

    /* Find the start of the password. */

    for (j = 4, i = 0; j && i < strlen(str); i++) {
        if (str[i] == '.') j--;
        }

    /* If there was no 4th dot then there can't be a password. */

    if (j) return (0);

    /* Send it to the validation routine */

    strcpy(tstr, crypt(&str[i], "ck"));
    if (strcmp(&tstr[2], base_password) == 0) return (1);
    return (0);
    }
