/*---------------------------------------------------------------------------
--
--  pec.c - Functions to handle Probalistic Error Checking in Project Checker
--
--  Copyright (c) 1992, 1993, 1994, 1995
--                University of Southern California, Los Angeles, CA
--                All rights reserved.
--
--  Redistribution and use of this software in source or binary form is
--  permitted given that the above copyright notice and this paragraph are
--  duplicated in all such forms and that any documentation and material
--  related to such distribution and use acknowledge that the software
--  was developed by University of Southern California.
--
--  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
--  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMIT TO, THE IMPLIED
--  WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
--  THE UNIVERSITY AND THE AUTHORS ASSUME NO RESPONSIBILITY FOR DAMAGES
--  OF ANY KIND RESULTING FROM THE USE OF THIS SOFTWARE.
--
--  Please contact Dr. Peter Danzig of Computer Science Department at
--  danzig@usc.edu for further information.
--
--
=============================================================================
--                             UPDATE HISTORY                              --
--                                                                         --
-- VER      Date         Explanation of Changes              Author        --
-----------------------------------------------------------------------------
--  0   | 2/94    | Initial version                 | Charles Chan - USC   --
--  1   | 5/95    | Minor editing                   | Steve Miller - USC   --
---------------------------------------------------------------------------*/

#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include "checker.h"

#define NextVictim(x) x = 1 + (int) (random () % pecIntervalNum)
#define PecAddQueryCount(ptr) if (ptr) (ptr)->queryCount++

void LogMessage();
char * _itoc();

extern char pecResultFileName[100];
extern u_long pecImmunePeriod;
extern int pecIntervalNum;
extern u_long pecTestPeriod;
extern time_t current_time;
extern PEC_LIST_TYPE  *pecList;
extern PEC_LIST_TYPE  *pecImmuneList;

PEC_LIST_TYPE *PecListInsert();
PEC_LIST_TYPE *PecFindServer();
void PecListRemove();
void ImmuneVictim();
void PecListMaintenance();
int  PecPrintList();


/*------------------------------------------------------------------------
--
--  int PecControl()
--     the entry point of PEC
--
--
--  GIVEN          : ip address of the resolver, and its particular
--                   query question, to be investigated
--  GLOBAL USED    : pecList, pecImmuneList
--  RETURNS        : 1 is the given revolver is captured by PEC, 0 
--                   otherwise
--  POST CONDITIONS: the post conditions of all other pec functions
--
--  COMMENTS:
--      You adjusted the pec algorithm, thus the way how pec works in
--      this function
--
-------------------------------------------------------------------------*/
int
PecControl(ip, name)
u_long  ip;
char   *name;
    {
    PEC_LIST_TYPE  *ptr;
    PEC_LIST_TYPE   newVictim;
    static int      count = 0;
    static int      victim = 5;
   
    
    /* first, immune any investigating resolver if its investigating
       period is over */
    
    PecListMaintenance();
        
    /* find out if the resolver is being investigated */

    ptr = PecFindServer(ip, pecList);
       
    /* if the resolver is being investigating...*/

    if (ptr) {

        /* also if the same query name */

        if ( !strcmp(name, ptr->query)) {
            PecAddQueryCount(ptr);
            return (ptr->queryCount);
            }
      
        /* the resolver is being investigated already but different
           query, let it go */

        return 0; 
        }

    /* if the resolver is not being investigated, check if it was
       investigated before */ 

    ptr = PecFindServer(ip, pecImmuneList);

    /* if the resolver have been investigated, let it go */

    if (ptr) return 0;
    count++;

    /* if this packet is the "victim"th packet in  total of 
       pecIntervalNum packets, it is the new victim */

    if (count == victim) {

        LogMessage("victim: %s query %s",_itoc((char *)&ip),name);

        /* add the poor victim to the pecList */

        bzero(newVictim, sizeof(PEC_LIST_TYPE));
        newVictim.queryCount = 1;
        newVictim.startTime = current_time;
        newVictim.resolver = ip;
        strcpy(newVictim.query, name);
        PecListInsert(&newVictim, &pecList);
        return 1;
        }

    /* adjust counter and choose the next victim from pecIntervalNum 
       packets */

    if (count > pecIntervalNum) {
        NextVictim(victim);
        count = 0;
        }

    return 0;
    }


/*------------------------------------------------------------------------
--
--  PEC_LIST_TYPE* PecListInsert():
--     Insert an element of PEC_LIST_TYPE to a list.
--
--  GIVEN          : element to insert, the list to be inserted to
--  GLOBAL USED    : none
--  RETURNS        : pointer to newly added element if successes, else NULL
--  POST CONDITIONS: new element added
--
--  COMMENTS:
--      None
--
-------------------------------------------------------------------------*/
PEC_LIST_TYPE *
PecListInsert(victim, list)
PEC_LIST_TYPE  *victim;
PEC_LIST_TYPE **list;
    {
    PEC_LIST_TYPE  *new;

    /* allocate new space */

    new = (PEC_LIST_TYPE *) malloc(sizeof(PEC_LIST_TYPE));
 
    if (new) {

        memcpy(new, victim, sizeof(PEC_LIST_TYPE));

        /* add the new one to the beginning of the list */

        new->next = *list;
        new->previous = NULL;
        if (*list) (*list)->previous = new;
        *list = new;

        return new;
        }

    return NULL;
    }


/*------------------------------------------------------------------------
--
--  PEC_LIST_TYPE* PecFindServer():
--     Locates the given resolver ip in the given list
--
--  GIVEN          : ip address og the resolver, the list to search
--  GLOBAL USED    : none
--  RETURNS        : pointer to the resolver record if found, else NULL
--  POST CONDITIONS: None
--
--  COMMENTS:
--      None
--
-------------------------------------------------------------------------*/
PEC_LIST_TYPE  *
PecFindServer(ip, list)
u_long          ip;
PEC_LIST_TYPE  *list;
    {
    PEC_LIST_TYPE  *ptr;
    
    ptr = list;

    while (ptr) {
        if (ptr->resolver == ip) return ptr;
        ptr = ptr->next;
        }

    return NULL;
    }


/*------------------------------------------------------------------------
--
--  void PecListRemove():
--     Removes an element from the given list pointed by the given pointer
--
--  GIVEN          : poiter to the element be removed, the list 
--  GLOBAL USED    : none
--  RETURNS        : none
--  POST CONDITIONS: the element is removed, the space occupied is freed
--
--  COMMENTS:
--      None
--
-------------------------------------------------------------------------*/
void
PecListRemove(ptr, list)
PEC_LIST_TYPE  *ptr;
PEC_LIST_TYPE **list;
    {

    if (ptr && *list) {

        /* the first element in the list */

        if (ptr == *list) {
            *list = ptr->next;
            if (*list)
            (*list)->previous = NULL;
            }

        else {
    
            /* the last element in the list */
    
            if (ptr->next == NULL)
                (ptr->previous)->next = NULL;

            /* the element in between */

            else {
                (ptr->previous)->next = ptr->next;
                (ptr->next)->previous = ptr->previous;
                }
            }

        free(ptr);
        }
    }


/*------------------------------------------------------------------------
--
--  void ImmuneVictim():
--     Removes an element from pecList to pecImmunelist 
--
--  GIVEN          : poiter to the element be removed to pecImmuneList
--  GLOBAL USED    : pecList, pecImmuneList
--  RETURNS        : none
--  POST CONDITIONS: the element is removed from pecList, and it is added
--                   to pecImmuneList
--
--  COMMENTS:
--      None
--
-------------------------------------------------------------------------*/
void
ImmuneVictim(pecListPtr)
PEC_LIST_TYPE  *pecListPtr;
    {

    if (pecListPtr) {

        /* assign the ending investigating time */

        pecListPtr->endTime = current_time;
        PecListInsert(pecListPtr, &pecImmuneList);
        PecListRemove(pecListPtr, &pecList);
        }
    }


/*------------------------------------------------------------------------
--
--  void PecListMaintenance()
--     Checks if it is time to stop investigating resolvers in pecList,
--     if so, immunes them.  Checks if it is time to remove any investigate
--     resolvers from pecImmuneList, if so remove them 
--
--  GIVEN          : none
--  GLOBAL USED    : pecList, pecImmuneList
--  RETURNS        : none
--  POST CONDITIONS: ImmuneVictim() is called, thus inherents the post
--                   conditions of the function.
--                   Investigated resolvers removed from pecImmuneList if
--                   their time expired
--
--  COMMENTS:
--      None
--
-------------------------------------------------------------------------*/
void 
PecListMaintenance()
    {
    PEC_LIST_TYPE  *list;
    static FILE *fp;
    static char init = 0;
    struct tm      *tmRec;

    if (!init) {

        init = 1;

        if ((fp = fopen(pecResultFileName, "a+")) == NULL)
            LogMessage("checker cannot open PEC result file");

        else
            { 
            fprintf(fp, "\n\n                 -=-  PEC Investigated");
            fprintf(fp, " Resolver List  -=-\n\n");
            fprintf(fp, "  Resolver            Start              ");
            fprintf(fp, "End            Tested      Query\n");
            fprintf(fp, "  IP Address          Time               Time");
            fprintf(fp, "           Time(mins)  Count\n");   
            fprintf(fp, "==============================================");
            fprintf(fp, "=================================\n");
            fflush(fp);
            } 
        }

    /* check the investigating resolver list */

    list = pecList;

    while (list) {

        /* update result report file */

        if (current_time - list->startTime > pecTestPeriod) {
            list->endTime = current_time;
            list->endTime = current_time;

            if (fp != NULL) { 

                tmRec = localtime(&(list->startTime));

                fprintf(fp, "%-15s  %02d:%02d:%02d-%02d/%02d/%02d ",
                    _itoc((char *) &list->resolver),
                    tmRec->tm_hour, tmRec->tm_min, tmRec->tm_sec,
                    tmRec->tm_mon+1, tmRec->tm_mday, tmRec->tm_year);

                tmRec = localtime(&(list->endTime));

                fprintf(fp,"   %02d:%02d:%02d-%02d/%02d/%02d %6ld %7d\n",
                    tmRec->tm_hour, tmRec->tm_min, tmRec->tm_sec,
                    tmRec->tm_mon+1, tmRec->tm_mday, tmRec->tm_year,
                    (list->endTime - list->startTime)/60,list->queryCount);

                fprintf(fp,"                 Query for: %s\n\n",list->query);
                fflush(fp);
                }

            /* move the resolver to the immune list */

            ImmuneVictim(list);
            }
        list = list->next;
        }

    /* check the investigated resolver list */

    list = pecImmuneList;
    while (list) {
        if (current_time - list->endTime > pecImmunePeriod)
            PecListRemove(list, &pecImmuneList);
        list = list->next;
        }

    }


/*------------------------------------------------------------------------
--
--  int PecPrintList()
--     print out the current status of the investigating or investigated
--     resolvers to the given file pointer
--
--
--  GIVEN          : the list to print -- pecList or pecImmuneList, the 
--                   file point that the result be printed to
--  GLOBAL USED    : pecList, pecImmuneList
--  RETURNS        : none
--  POST CONDITIONS: the status of the list is printed to fp
--
--  COMMENTS:
--      None
--
-------------------------------------------------------------------------*/
int
PecPrintList(list, fp)
PEC_LIST_TYPE  *list;
FILE           *fp;
    {
    struct tm      *tmRec;
    PEC_LIST_TYPE *whichList;

    whichList = list;

    /* print out header lines */

    if (list == pecList) {
        fprintf(fp, "\n\n                -=-  PEC Investigating Resolver List  -=-\n\n");
        fprintf(fp, "  Resolver            Started        Query     Query\n");
        fprintf(fp, "  IP Address          Time           Count     Question\n");
        }

    else {
        fprintf(fp, "\n\n                 -=-  PEC Investigated Resolver List  -=-\n\n");
        fprintf(fp, "  Resolver            Start              End            Tested      Query\n");
        fprintf(fp, "  IP Address          Time               Time           Time(mins)  Count\n");
        }

    fprintf(fp, "===============================================================================\n");
    
    /* print out empty list message if nothing in the list */ 

    if (!list && whichList == pecList) 
        fprintf(fp,"\n      < no resolver is being investigated > \n\n");   
   
    if (!list && whichList != pecList) 
        fprintf(fp,"\n      < no resolver is done with investigation > \n\n"); 

    /* print out detail info */

    while (list) {

        tmRec = localtime(&(list->startTime));

        fprintf(fp, "%-15s  %02d:%02d:%02d-%02d/%02d/%02d ",
            _itoc((char *) &list->resolver),
            tmRec->tm_hour, tmRec->tm_min, tmRec->tm_sec,
            tmRec->tm_mon+1, tmRec->tm_mday, tmRec->tm_year);

        if (whichList == pecList)
            fprintf(fp,"%7d     %s\n",list->queryCount, list->query);

        else {
            tmRec = localtime(&(list->endTime));
            fprintf(fp,"   %02d:%02d:%02d-%02d/%02d/%02d %6ld %7d\n",
                tmRec->tm_hour, tmRec->tm_min, tmRec->tm_sec,
                tmRec->tm_mon+1, tmRec->tm_mday, tmRec->tm_year,
                (list->endTime - list->startTime)/60,list->queryCount);
            fprintf(fp,"                 Query for: %s\n\n",list->query);
            }

        list = list->next;
        }

    fprintf(fp, "-------------------------------------------------------------------------------\n");
    fflush(fp);
    }


/*-----------------------------------------------------------------------
--
--  itoc - returns inet address in dot format given HEX address
--
-----------------------------------------------------------------------*/
char *
_itoc(str)
char  *str;
    {
    int             i;
    char            asc[8];
    static char     rlt[17];

    rlt[0] = '\0';

    for (i = 0; i < sizeof(u_long); i++) {
        sprintf(asc, "%d.", (u_char) str[i]);
        strcat(rlt, asc);
        }

    rlt[strlen(rlt) - 1] = '\0';
    return (rlt);
    }
